package dev.fyloz.trial.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.trial.colorrecipesexplorer.exception.model.EntityAlreadyExistsRestException
import dev.fyloz.trial.colorrecipesexplorer.model.*
import dev.fyloz.trial.colorrecipesexplorer.repository.MaterialRepository
import dev.fyloz.trial.colorrecipesexplorer.service.files.SimdutService
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Nested
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class MaterialServiceTest :
    AbstractExternalNamedModelServiceTest<Material, MaterialSaveDto, MaterialUpdateDto, MaterialService, MaterialRepository>() {
    override val repository: MaterialRepository = mock()
    private val simdutService: SimdutService = mock()
    private val recipeService: RecipeService = mock()
    private val mixService: MixService = mock()
    override val service: MaterialService =
        spy(MaterialServiceImpl(repository, simdutService, recipeService, mixService))

    override val entity: Material = material(id = 0L, name = "material")
    override val anotherEntity: Material = material(id = 1L, name = "another material")
    override val entityWithEntityName: Material = material(id = 2L, name = "material")
    override val entitySaveDto: MaterialSaveDto = spy(materialSaveDto())
    override val entityUpdateDto: MaterialUpdateDto = spy(materialUpdateDto(id = 0L))

    private val materialType = materialType()

    @AfterEach
    override fun afterEach() {
        reset(simdutService)
        super.afterEach()
    }

    // existsByMaterialType

    @Test
    fun `existsByMaterialType() returns true when a material with the given material type exists in the repository`() {
        whenever(repository.existsByMaterialType(materialType)).doReturn(true)

        val found = service.existsByMaterialType(materialType)

        assertTrue(found)
    }

    @Test
    fun `existsByMaterialType() returns false when no material with the given material type exists in the repository`() {
        whenever(repository.existsByMaterialType(materialType)).doReturn(false)

        val found = service.existsByMaterialType(materialType)

        assertFalse(found)
    }

    // hasSimdut()

    @Test
    fun `hasSimdut() returns false when simdutService_exists() returns false`() {
        whenever(simdutService.exists(entity)).doReturn(false)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.hasSimdut(entity.id!!)

        assertFalse(found)
    }

    @Test
    fun `hasSimdut() returns true when simdutService_exists() returns true`() {
        whenever(simdutService.exists(entity)).doReturn(true)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.hasSimdut(entity.id!!)

        assertTrue(found)
    }

    // getAllNotMixType()

    @Test
    fun `getAllNotMixType() returns a list containing every material that are not a mix type`() {
        val mixTypeMaterial = material(name = "mix type material", isMixType = true)
        val materialList = listOf(entity, mixTypeMaterial)

        doReturn(materialList).whenever(service).getAll()

        val found = service.getAllNotMixType()

        assertTrue(found.contains(entity))
        assertFalse(found.contains(mixTypeMaterial))
    }

    // save()

    @Test
    fun `save() throws EntityAlreadyExistsRestException when a material with the given name exists in the repository`() {
        doReturn(true).whenever(service).existsByName(entity.name)

        val exception = assertThrows<EntityAlreadyExistsRestException> { service.save(entity) }
        assertEquals(entity.name, exception.value)
    }

    @Test
    fun `save(dto) calls simdutService_write() with the saved entity`() {
        val mockMultipartFile = spy(MockMultipartFile("simdut", byteArrayOf()))
        val materialSaveDto = spy(materialSaveDto(simdutFile = mockMultipartFile))

        doReturn(false).whenever(mockMultipartFile).isEmpty
        doReturn(entity).whenever(materialSaveDto).toEntity()
        doReturn(entity).whenever(service).save(entity)

        service.save(materialSaveDto)

        verify(simdutService).write(entity, mockMultipartFile)
    }

    // update()

    @Test
    fun `update() throws EntityAlreadyExistsRestException when another material with the updated name exists in the repository`() {
        val material = material(id = 0L, name = "name")
        val anotherMaterial = material(id = 1L, name = "name")

        whenever(repository.findByName(material.name)).doReturn(anotherMaterial)
        doReturn(entity).whenever(service).getById(material.id!!)

        val exception = assertThrows<EntityAlreadyExistsRestException> { service.update(material) }
        assertEquals(material.name, exception.value)
    }

    // getAllForMixCreation()

    @Test
    fun `getAllForMixCreation() returns all normal materials and all mix type materials for the given recipe`() {
        val normalMaterial = material(id = 0L, isMixType = false)
        val mixTypeMaterial = material(id = 1L, isMixType = true)
        val anotherMixTypeMaterial = material(id = 2L, isMixType = true)
        val materials = listOf(normalMaterial, mixTypeMaterial, anotherMixTypeMaterial)
        val recipe =
            recipe(id = 0L, mixes = mutableListOf(mix(mixType = mixType(id = 0L, material = mixTypeMaterial))))

        whenever(recipeService.getById(recipe.id!!)).doReturn(recipe)
        doReturn(materials).whenever(service).getAll()

        val found = service.getAllForMixCreation(recipe.id!!)

        assertTrue(found contains normalMaterial)
        assertTrue(found contains mixTypeMaterial)
        assertFalse(found contains anotherMixTypeMaterial)
    }

    // getAllForMixUpdate()

    @Test
    fun `getAllForMixUpdate() returns all normal materials and all mix type materials for the recipe of the given mix without the mix type of the said mix`() {
        val normalMaterial = material(id = 0L, isMixType = false)
        val mixTypeMaterial = material(id = 1L, isMixType = true)
        val anotherMixTypeMaterial = material(id = 2L, isMixType = true)
        val materials = listOf(normalMaterial, mixTypeMaterial, anotherMixTypeMaterial)
        val recipe = recipe(id = 0L, mixes = mutableListOf(mix(mixType = mixType(material = mixTypeMaterial))))
        val mix = mix(id = 1L, recipe = recipe, mixType = mixType(material = anotherMixTypeMaterial))
        recipe.mixes.add(mix)

        whenever(mixService.getById(mix.id!!)).doReturn(mix)
        doReturn(materials).whenever(service).getAll()

        val found = service.getAllForMixUpdate(mix.id!!)

        assertTrue(found contains normalMaterial)
        assertTrue(found contains mixTypeMaterial)
        assertFalse(found contains anotherMixTypeMaterial)
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        val mockSimdutFile = MockMultipartFile("simdut", byteArrayOf(1, 2, 3, 4, 5))
        val materialUpdateDto = spy(materialUpdateDto(id = 0L, simdutFile = mockSimdutFile))

        doReturn(entity).whenever(service).getById(materialUpdateDto.id)
        doReturn(entity).whenever(service).update(any<Material>())
        doReturn(entity).whenever(materialUpdateDto).toEntity()

        service.update(materialUpdateDto)

        verify(simdutService).update(eq(mockSimdutFile), any())
    }

    /** Helper function to replace collections.in because the id is not considered in the equals function of Material while Thymeleaf is supported. */
    private infix fun Collection<Material>.contains(material: Material): Boolean =
        any { it.id == material.id }
}
